/*
   Convert a PPM image into the DjVu Color RLE format
   described in the csepdjvu(1) documentation
  
   Copyright (c) 2004 Scott Pakin <scott+pbm@pakin.org>

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <stdio.h>
#include <inttypes.h>
#include <netinet/in.h>

#include "ppm.h"
#include "ppmcmap.h"


/* Write a color and run length to the RLE file. */
static void 
write_rle (FILE *          const rlefile, 
           pixel           const color, 
           int             const tally,
           colorhash_table const colorhash,
           pixel           const transcolor)
{
  uint32_t rlevalue;         /* RLE-encoded {color, value} pair */

  if (!tally)
    return;
  rlevalue = (PPM_EQUAL (color, transcolor) ?
              0xFFF :
              ppm_lookupcolor (colorhash, &color));
  if (rlevalue == -1)
    pm_error ("internal error -- color not found in hash table");
  rlevalue = htonl ((rlevalue<<20) | tally);
  if (fwrite ((const void *) &rlevalue, sizeof(uint32_t), 1, rlefile) != 1)
    pm_perror ("failed to write to the RLE file");
}


/* Perform the bulk of the work. */
int main (int argc, char *argv[])
{
  FILE *ppmfile = stdin;     /* PPM file to convert */
  FILE * const rlefile = stdout;    /* Generated Color RLE file */
  int numcols, numrows;      /* Width and height in pixels of the PPM file */
  pixval maxval;             /* Maximum value for a color component */
  int format;                /* Original image type before conversion to PPM */
  pixel *ppmrow;             /* One row of the PPM file */
  colorhist_vector colormap; /* List of all of the colors used */
  int numcolors;             /* Number of unique colors in the color map */
  unsigned int postheaderpos;  /* File offset pointing just past the header */
  const char *usagemsg = "[-transparent <color>] infile.ppm [outfile.rle]";
  const char *transcolorname = "white";   /* Name of the transparent color */
  int pixeltally = 0;        /* Run length of prevpixel */
  int row, col;              /* Row and column loop variables */
  int i;
  colorhash_table colorhash; 
    /* Mapping from color to index into colormap[] */
  pixel transcolor;
    /* Color that should be considered transparent */


  /* Parse the command line. */
  ppm_init (&argc, argv);
  for (i=1; i<argc; i++)
    if (pm_keymatch (argv[i], "-help", 2))
      pm_usage (usagemsg);
    else
      if (pm_keymatch (argv[i], "-transparent", 2))
        if (i+1 < argc)
          transcolorname = argv[++i];
        else
          pm_usage (usagemsg);
      else
        if (ppmfile == stdin)
          ppmfile = pm_openr_seekable (argv[i]);

  /* Write an RLE header and color map. */
  ppm_readppminit (ppmfile, &numcols, &numrows, &maxval, &format);
  postheaderpos = pm_tell (ppmfile);
  transcolor = ppm_parsecolor (transcolorname, maxval);
  colormap = ppm_computecolorhist2 (ppmfile, numcols, numrows, maxval, format,
				    (maxval+1)*(maxval+1)*(maxval+1), &numcolors);
  if (numcolors > 0xFF0)
    pm_error ("too many colors; "
              "use pnmquant to reduce to no more than 4080 colors");
  colorhash = ppm_colorhisttocolorhash (colormap, numcolors);
  fprintf (rlefile, "R6\n");
  fprintf (rlefile, "%d %d %d\n", numcols, numrows, numcolors);
  for (i=0; i<numcolors; i++) {
    pixel scaledcolor;           /* Pixel scaled to 255 colors */

    PPM_DEPTH (scaledcolor, colormap[i].color, maxval, 255);
    fputc (scaledcolor.r, rlefile);
    fputc (scaledcolor.g, rlefile);
    fputc (scaledcolor.b, rlefile);
  }

  /* Write the RLE data. */
  pm_seek (ppmfile, postheaderpos);
  ppmrow = ppm_allocrow (numcols);
  for (row=0; row<numrows; row++) {
    pixel prevpixel;        /* Previous pixel seen */

    ppm_readppmrow (ppmfile, ppmrow, numcols, maxval, format);
    prevpixel = ppmrow[0];
    pixeltally = 0;
    for (col=0; col<numcols; col++) {
      pixel newpixel = ppmrow[col];      /* Current pixel color */

      if (PPM_EQUAL (newpixel, prevpixel))
        pixeltally++;
      else {
        write_rle (rlefile, prevpixel, pixeltally, colorhash, transcolor);
        pixeltally = 1;
        prevpixel = newpixel;
      }
      if (pixeltally >= (1<<20)-1) {
        write_rle (rlefile, prevpixel, pixeltally, colorhash, transcolor);
        pixeltally = 1;
      }
    }
    write_rle (rlefile, prevpixel, pixeltally, colorhash, transcolor);
  }

  /* Finish up cleanly. */
  ppm_freerow (ppmrow);
  ppm_freecolorhash (colorhash);
  ppm_freecolorhist (colormap);
  if (ppmfile != stdin)
    pm_close (ppmfile);
  return 0;
}

